/* c_count: given a list of C/C++/Java files on the command line,
   count the SLOC in each one.  SLOC = physical, non-comment lines.
   This program knows about C++ and C comments (and how they interact),
   and correctly ignores comment markers inside strings.

   (C) Copyright 2000 David A. Wheeler
   Covered by the GNU GPL, version 2.
   This version 2000-Oct-5.
   This is part of "SLOCCount"; see the documentation for details.

   Usage: Use in one of the following ways:
     c_count                      # As filter
     c_count list_of_files        # Counts for each file.
     c_count -f fl                # Counts the files listed in "fl".

*/

#include <stdio.h>
#include <string.h>

/* Modes */
#define NORMAL 0
#define INSTRING 1
#define INCOMMENT 2

/* Types of comments: */
#define ANSIC_STYLE 0
#define CPP_STYLE 1


/* Globals */
long total_sloc;

static warn_embedded_newlines = 0;

int peek() {
 int c = getchar();
 ungetc(c, stdin);
 return c;
}

int ispeek(int c) {
 if (c == peek()) {return 1;}
 return 0;
}

long line_number;

int getachar() {
/* Like getchar(), but keep track of line number. */
 static last_char_was_newline = 0;
 int c;

 c = getchar(); 
 if (last_char_was_newline) line_number++;
 if (c == '\n') last_char_was_newline=1;
 else           last_char_was_newline=0;
 return c;
}


long sloc_count(char *filename) {
 /* Count the sloc in the program in stdin. */

 long sloc = 0;

 int sawchar = 0;                /* Did you see a character on this line? */
 int c;
 int mode = NORMAL;              /* NORMAL, INSTRING, or INCOMMENT */
 int comment_type = ANSIC_STYLE; /* ANSIC_STYLE or CPP_STYLE */
 

 /* The following implements a state machine with transitions; the
    main state is "mode" and "comment_type", the transitions are
    triggered by characters input. */

 while ( (c = getachar()) != EOF) {
   if      (mode == NORMAL) {
     if (c == '"') {sawchar=1; mode = INSTRING;}
     else if (c == '\'') {  /* Consume single-character 'xxxx' values */
       sawchar=1;
       c = getachar();
       if (c == '\\') c = getachar();
       do {
         c = getachar();
       } while ((c != '\'') && (c != '\n') & (c != EOF));
     } else if ((c == '/') && ispeek('*')) {
          c = getachar();
          mode = INCOMMENT;
          comment_type = ANSIC_STYLE;
     } else if ((c == '/') && ispeek('/')) {
          c = getachar();
          mode = INCOMMENT;
          comment_type = CPP_STYLE;
     } else if (!isspace(c)) {sawchar = 1;}
   } else if (mode == INSTRING) {
     /* We only count string lines with non-whitespace -- this is to
        gracefully handle syntactically invalid programs.
        You could argue that multiline strings with whitespace are
        still executable and should be counted. */
     if (!isspace(c)) sawchar = 1;
     if (c == '"') {mode = NORMAL;}
     else if ((c == '\\') && (ispeek('\"') || ispeek('\\'))) {c = getachar();}
     else if ((c == '\\') && ispeek('\n')) {c = getachar();}
     else if ((c == '\n') && warn_embedded_newlines) {
       /* We found a bare newline in a string without preceding backslash. */
       fprintf(stderr, "c_count WARNING - newline in string, line %ld, file %s\n", line_number, filename);
       /* We COULD warn & reset mode to "Normal", but lots of code does this,
          so we'll just depend on the warning for ending the program
          in a string to catch syntactically erroneous programs. */
     }
   } else {  /* INCOMMENT mode */
     if ((c == '\n') && (comment_type == CPP_STYLE)) { mode = NORMAL;}
     if ((comment_type == ANSIC_STYLE) && (c == '*') &&
          ispeek('/')) { c= getachar(); mode = NORMAL;}
   }
   if (c == '\n') {
     if (sawchar) sloc++;
     sawchar = 0;
   }
 }
 /* We're done with the file.  Handle EOF-without-EOL. */
 if (sawchar) sloc++;
 sawchar = 0;
 if ((mode == INCOMMENT) && (comment_type == CPP_STYLE)) { mode = NORMAL;}

 if (mode == INCOMMENT) {
     fprintf(stderr, "c_count ERROR - terminated in comment in %s\n", filename);
 } else if (mode == INSTRING) {
     fprintf(stderr, "c_count ERROR - terminated in string in %s\n", filename);
 }

 return sloc;
}


void count_file(char *filename) {
  long sloc;

  fclose(stdin);
  freopen(filename, "r", stdin);
  line_number = 1;
  sloc = sloc_count(filename);
  total_sloc += sloc;
  printf("%ld %s\n", sloc, filename);
}

char *read_a_line(FILE *file) {
 /* Read a line in, and return a malloc'ed buffer with the line contents.
    Any newline at the end is stripped.
    If there's nothing left to read, returns NULL. */

 /* We'll create a monstrously long buffer to make life easy for us: */
 char buffer[10000];
 char *returnval;
 char *newlinepos;

 returnval = fgets(buffer, sizeof(buffer), file);
 if (returnval) {
   newlinepos = buffer + strlen(buffer) - 1;
   if (*newlinepos == '\n') {*newlinepos = '\0';};
   return strdup(buffer);
 } else {
   return NULL;
 }
}


main(int argc, char *argv[]) {
 long sloc;
 int i;
 FILE *file_list;
 char *s;

 total_sloc = 0;
 line_number = 1;

 if (argc <= 1) {
   sloc = sloc_count("-");
   printf("%ld %s\n", sloc, "-");
 } else if ((argc == 3) && (!strcmp(argv[1], "-f"))) {
   file_list = fopen(argv[2], "r");
   if (file_list) {
     while (s = read_a_line(file_list)) {
       count_file(s);
       free(s);
     }
   }
 } else {
   for (i=1; i < argc; i++) { count_file(argv[i]); }
 }
 printf("Total:\n");
 printf("%ld\n", total_sloc);
}

